@@ -1,10 +1,12 @@
 use rand::Rng;
 use rnp::{
-    PingClientConfig, PingResultProcessorConfig, PingWorkerConfig, PingWorkerSchedulerConfig,
-    PortRangeList, RnpCoreConfig, RnpSupportedProtocol,
+    parse_ping_target, PingClientConfig, PingResultProcessorCommonConfig, PingResultProcessorConfig, PingWorkerConfig, PingWorkerSchedulerConfig,
+    PortRangeList, RnpPingRunnerConfig, RnpSupportedProtocol,
 };
 use std::net::{IpAddr, Ipv4Addr, Ipv6Addr, SocketAddr};
 use std::path::PathBuf;
+
+use std::sync::{Arc, Mutex};
 use std::time::Duration;
 use structopt::StructOpt;
 
@@ -12,33 +14,30 @@ use structopt::StructOpt;
 #[structopt(name = rnp::RNP_NAME, author = rnp::RNP_AUTHOR, about = rnp::RNP_ABOUT)]
 pub struct RnpCliOptions {
     #[structopt(flatten)]
-    common_options: RnpCliCommonOptions,
+    pub common_options: RnpCliCommonOptions,
 
     #[structopt(flatten)]
-    output_options: RnpCliOutputOptions,
+    pub ping_common_options: RnpCliPingCommonOptions,
 
     #[structopt(flatten)]
-    quic_options: RnpCliQuicPingOptions,
+    pub output_options: RnpCliOutputOptions,
+
+    #[structopt(flatten)]
+    pub quic_options: RnpCliQuicPingOptions,
 }
 
 #[derive(Debug, StructOpt, PartialEq)]
 pub struct RnpCliCommonOptions {
-    pub target: SocketAddr,
-
-    #[structopt(
-        short = "m",
-        long = "mode",
-        default_value = "TCP",
-        help = "Specify protocol to use."
-    )]
+    #[structopt(short = "m", long = "mode", default_value = "TCP", help = "Specify protocol to use.")]
     pub protocol: RnpSupportedProtocol,
 
-    #[structopt(
-        short = "s",
-        long = "src-ip",
-        default_value = "0.0.0.0",
-        help = "Source IP address."
-    )]
+    #[structopt(parse(try_from_str = parse_ping_target), help = "Target endpoint. For IPv6, please use [] to wrap the address, such as [::1]:80.")]
+    pub target: SocketAddr,
+}
+
+#[derive(Debug, StructOpt, PartialEq)]
+pub struct RnpCliPingCommonOptions {
+    #[structopt(short = "s", long = "src-ip", default_value = "0.0.0.0", help = "Source IP address.")]
     pub source_ip: IpAddr,
 
     #[structopt(
@@ -57,20 +56,10 @@ pub struct RnpCliCommonOptions {
     #[structopt(long = "warmup", default_value = "0", help = "Warm up ping count.")]
     pub warmup_count: u32,
 
-    #[structopt(
-        short = "w",
-        long = "timeout",
-        default_value = "2000",
-        help = "Wait time for each ping in milliseconds."
-    )]
+    #[structopt(short = "w", long = "timeout", default_value = "2000", help = "Wait time for each ping in milliseconds.")]
     pub wait_timeout_in_ms: u32,
 
-    #[structopt(
-        short = "i",
-        long = "interval",
-        default_value = "1000",
-        help = "Sleep between each ping in milliseconds."
-    )]
+    #[structopt(short = "i", long = "interval", default_value = "1000", help = "Sleep between each ping in milliseconds.")]
     pub ping_interval_in_ms: u32,
 
     #[structopt(long = "ttl", help = "Time to live.")]
@@ -84,59 +73,49 @@ pub struct RnpCliCommonOptions {
     pub check_disconnect: bool,
 
     #[structopt(
-        short = "p",
-        long = "parallel",
-        default_value = "1",
-        help = "Count of pings running in parallel."
+        long = "wait-before-disconnect",
+        default_value = "0",
+        help = "Wait before disconnect in milliseconds. Only works when check-disconnect is enabled."
     )]
+    pub wait_before_disconnect_in_ms: u64,
+
+    #[structopt(
+        long = "disconnect-timeout",
+        alias = "dt",
+        default_value = "2000",
+        help = "After initiate disconnect, wait for specified milliseconds to check if disconnect can be successfully completed. [alias: --dt]"
+    )]
+    pub disconnect_timeout_in_ms: u64,
+
+    #[structopt(short = "p", long = "parallel", default_value = "1", help = "Count of pings running in parallel.")]
     pub parallel_ping_count: u32,
+
+    #[structopt(long, help = "Exit as soon as a ping failed and return a non-zero error code.")]
+    pub exit_on_fail: bool,
 }
 
 #[derive(Debug, StructOpt, PartialEq)]
 pub struct RnpCliOutputOptions {
     #[structopt(
         short = "q",
-        long,
-        help = "Don't log each ping result to console. Summary and other things will still be written to console."
+        parse(from_occurrences),
+        help = "Quiet mode. -q = Don't output each ping result; -qq = Don't output final summaries; -qqq = Don't output anything."
     )]
-    pub no_console_log: bool,
+    pub quiet_level: i32,
 
-    #[structopt(
-        long = "log-csv",
-        alias = "oc",
-        parse(from_os_str),
-        help = "Log ping results a csv file. [alias: --oc]"
-    )]
+    #[structopt(long = "log-csv", alias = "oc", parse(from_os_str), help = "Log ping results a csv file. [alias: --oc]")]
     pub csv_log_path: Option<PathBuf>,
 
-    #[structopt(
-        long = "log-json",
-        alias = "oj",
-        parse(from_os_str),
-        help = "Log ping results to a json file. [alias: --oj]"
-    )]
+    #[structopt(long = "log-json", alias = "oj", parse(from_os_str), help = "Log ping results to a json file. [alias: --oj]")]
     pub json_log_path: Option<PathBuf>,
 
-    #[structopt(
-        short = "o",
-        long = "log-text",
-        parse(from_os_str),
-        help = "Log ping results to a text file."
-    )]
+    #[structopt(short = "o", long = "log-text", parse(from_os_str), help = "Log ping results to a text file.")]
     pub text_log_path: Option<PathBuf>,
 
-    #[structopt(
-        short = "r",
-        long,
-        help = "Show ping result scatter map after ping is done."
-    )]
+    #[structopt(short = "r", long, help = "Show ping result scatter map after ping is done.")]
     pub show_result_scatter: bool,
 
-    #[structopt(
-        short = "l",
-        long,
-        help = "Show latency (round trip time) scatter map after ping is done."
-    )]
+    #[structopt(short = "l", long, help = "Show latency (round trip time) scatter map after ping is done.")]
     pub show_latency_scatter: bool,
 
     #[structopt(
@@ -149,11 +128,8 @@ pub struct RnpCliOutputOptions {
 }
 
 #[derive(Debug, StructOpt, PartialEq)]
-struct RnpCliQuicPingOptions {
-    #[structopt(
-        long,
-        help = "Specify the server name in the QUIC pings. Example: localhost."
-    )]
+pub struct RnpCliQuicPingOptions {
+    #[structopt(long, help = "Specify the server name in the QUIC pings. Example: localhost.")]
     pub server_name: Option<String>,
 
     #[structopt(
@@ -178,42 +154,32 @@ struct RnpCliQuicPingOptions {
 
 impl RnpCliOptions {
     pub fn prepare_to_use(&mut self) {
-        self.common_options.prepare_to_use();
+        self.ping_common_options.prepare_to_use(&self.common_options.target);
 
         if let Some(latency_buckets) = &mut self.output_options.latency_buckets {
             tracing::debug!("Latency bucket set to 0. Use default one.");
-            if latency_buckets.len() == 0
-                || (latency_buckets.len() == 1 && latency_buckets[0] == 0.0)
-            {
+            if latency_buckets.len() == 0 || (latency_buckets.len() == 1 && latency_buckets[0] == 0.0) {
                 *latency_buckets = vec![0.1, 0.5, 1.0, 5.0, 10.0, 30.0, 50.0, 100.0, 300.0, 500.0];
             }
         }
     }
 
-    pub fn to_rnp_core_config(&self) -> RnpCoreConfig {
-        let mut config = RnpCoreConfig {
+    pub fn to_ping_runner_config(&self) -> RnpPingRunnerConfig {
+        let mut config = RnpPingRunnerConfig {
             worker_config: PingWorkerConfig {
                 protocol: self.common_options.protocol.clone(),
                 target: self.common_options.target,
-                source_ip: self.common_options.source_ip,
-                ping_interval: Duration::from_millis(
-                    self.common_options.ping_interval_in_ms.into(),
-                ),
+                source_ip: self.ping_common_options.source_ip,
+                ping_interval: Duration::from_millis(self.ping_common_options.ping_interval_in_ms.into()),
                 ping_client_config: PingClientConfig {
-                    wait_timeout: Duration::from_millis(
-                        self.common_options.wait_timeout_in_ms.into(),
-                    ),
-                    time_to_live: self.common_options.time_to_live,
-                    check_disconnect: self.common_options.check_disconnect,
-                    server_name: self
-                        .quic_options
-                        .server_name
-                        .as_ref()
-                        .and_then(|s| Some(s.to_string())),
+                    wait_timeout: Duration::from_millis(self.ping_common_options.wait_timeout_in_ms.into()),
+                    time_to_live: self.ping_common_options.time_to_live,
+                    check_disconnect: self.ping_common_options.check_disconnect,
+                    wait_before_disconnect: Duration::from_millis(self.ping_common_options.wait_before_disconnect_in_ms),
+                    disconnect_timeout: Duration::from_millis(self.ping_common_options.disconnect_timeout_in_ms),
+                    server_name: self.quic_options.server_name.as_ref().and_then(|s| Some(s.to_string())),
                     log_tls_key: self.quic_options.log_tls_key,
-                    alpn_protocol: if self.quic_options.alpn_protocol.to_uppercase()
-                        != String::from("NONE")
-                    {
+                    alpn_protocol: if self.quic_options.alpn_protocol.to_uppercase() != String::from("NONE") {
                         Some(self.quic_options.alpn_protocol.clone())
                     } else {
                         None
@@ -222,46 +188,40 @@ impl RnpCliOptions {
                 },
             },
             worker_scheduler_config: PingWorkerSchedulerConfig {
-                source_ports: self.common_options.source_ports.as_ref().unwrap().clone(),
+                source_ports: self.ping_common_options.source_ports.as_ref().unwrap().clone(),
                 ping_count: None,
-                warmup_count: self.common_options.warmup_count,
-                parallel_ping_count: self.common_options.parallel_ping_count,
+                warmup_count: self.ping_common_options.warmup_count,
+                parallel_ping_count: self.ping_common_options.parallel_ping_count,
             },
             result_processor_config: PingResultProcessorConfig {
-                no_console_log: self.output_options.no_console_log,
+                common_config: PingResultProcessorCommonConfig { quiet_level: self.output_options.quiet_level },
+                exit_on_fail: self.ping_common_options.exit_on_fail,
+                exit_failure_reason: if self.ping_common_options.exit_on_fail { Some(Arc::new(Mutex::new(None))) } else { None },
                 csv_log_path: self.output_options.csv_log_path.clone(),
                 json_log_path: self.output_options.json_log_path.clone(),
                 text_log_path: self.output_options.text_log_path.clone(),
                 show_result_scatter: self.output_options.show_result_scatter,
                 show_latency_scatter: self.output_options.show_latency_scatter,
-                latency_buckets: self
-                    .output_options
-                    .latency_buckets
-                    .as_ref()
-                    .and_then(|buckets| Some(buckets.clone())),
+                latency_buckets: self.output_options.latency_buckets.as_ref().and_then(|buckets| Some(buckets.clone())),
             },
             external_ping_client_factory: None,
             extra_ping_result_processors: vec![],
         };
 
-        if !self.common_options.ping_until_stopped {
-            config.worker_scheduler_config.ping_count = Some(self.common_options.ping_count);
+        if !self.ping_common_options.ping_until_stopped {
+            config.worker_scheduler_config.ping_count = Some(self.ping_common_options.ping_count);
         }
 
         return config;
     }
 }
 
-impl RnpCliCommonOptions {
-    pub fn prepare_to_use(&mut self) {
-        if self.target.is_ipv4() != self.source_ip.is_ipv4() {
+impl RnpCliPingCommonOptions {
+    pub fn prepare_to_use(&mut self, target: &SocketAddr) {
+        if target.is_ipv4() != self.source_ip.is_ipv4() {
             match &self.source_ip {
-                IpAddr::V4(source_ip_v4) if *source_ip_v4 == Ipv4Addr::UNSPECIFIED => {
-                    self.source_ip = IpAddr::V6(Ipv6Addr::UNSPECIFIED)
-                }
-                IpAddr::V6(source_ip_v6) if *source_ip_v6 == Ipv6Addr::UNSPECIFIED => {
-                    self.source_ip = IpAddr::V4(Ipv4Addr::UNSPECIFIED)
-                }
+                IpAddr::V4(source_ip_v4) if *source_ip_v4 == Ipv4Addr::UNSPECIFIED => self.source_ip = IpAddr::V6(Ipv6Addr::UNSPECIFIED),
+                IpAddr::V6(source_ip_v6) if *source_ip_v6 == Ipv6Addr::UNSPECIFIED => self.source_ip = IpAddr::V4(Ipv4Addr::UNSPECIFIED),
                 _ => panic!("Source IP and Target IP are not both IPv4 or IPv6!"),
             }
         }
@@ -269,9 +229,7 @@ impl RnpCliCommonOptions {
         if self.source_ports.is_none() {
             let range_start = rand::thread_rng().gen_range(10000..30000);
             let range_end = range_start + 2000;
-            self.source_ports = Some(PortRangeList {
-                ranges: vec![(range_start..=range_end)],
-            });
+            self.source_ports = Some(PortRangeList { ranges: vec![(range_start..=range_end)] });
         }
 
         if !self.ping_until_stopped && self.ping_count < 1 {
@@ -279,11 +237,7 @@ impl RnpCliCommonOptions {
             self.ping_count = 1;
         }
 
-        let available_source_port_count = self
-            .source_ports
-            .as_ref()
-            .unwrap()
-            .calculate_total_port_count();
+        let available_source_port_count = self.source_ports.as_ref().unwrap().calculate_total_port_count();
         if self.parallel_ping_count > available_source_port_count as u32 {
             tracing::warn!(
                 "Parallel ping count ({}) is larger than available source port count ({}), to avoid port conflict reducing parallel ping count down to the same as available source port count.",
@@ -303,9 +257,10 @@ impl RnpCliCommonOptions {
 #[cfg(test)]
 mod tests {
     use super::*;
+    use pretty_assertions::assert_eq;
     use rnp::{
-        PingClientConfig, PingResultProcessorConfig, PingWorkerConfig, PingWorkerSchedulerConfig,
-        RnpCoreConfig, RnpSupportedProtocol,
+        PingClientConfig, PingResultProcessorCommonConfig, PingResultProcessorConfig, PingWorkerConfig, PingWorkerSchedulerConfig,
+        RnpPingRunnerConfig, RnpSupportedProtocol, RNP_QUIET_LEVEL_NONE, RNP_QUIET_LEVEL_NO_OUTPUT, RNP_QUIET_LEVEL_NO_PING_RESULT,
     };
     use std::path::PathBuf;
     use std::time::Duration;
@@ -315,9 +270,8 @@ mod tests {
     fn parsing_default_options_should_work() {
         assert_eq!(
             RnpCliOptions {
-                common_options: RnpCliCommonOptions {
-                    target: "10.0.0.1:443".parse().unwrap(),
-                    protocol: RnpSupportedProtocol::TCP,
+                common_options: RnpCliCommonOptions { target: "10.0.0.1:443".parse().unwrap(), protocol: RnpSupportedProtocol::TCP },
+                ping_common_options: RnpCliPingCommonOptions {
                     source_ip: "0.0.0.0".parse().unwrap(),
                     source_ports: None,
                     ping_count: 4,
@@ -327,7 +281,10 @@ mod tests {
                     ping_interval_in_ms: 1000,
                     time_to_live: None,
                     check_disconnect: false,
+                    wait_before_disconnect_in_ms: 0,
+                    disconnect_timeout_in_ms: 2000,
                     parallel_ping_count: 1,
+                    exit_on_fail: false,
                 },
                 quic_options: RnpCliQuicPingOptions {
                     server_name: None,
@@ -336,7 +293,7 @@ mod tests {
                     use_timer_rtt: false,
                 },
                 output_options: RnpCliOutputOptions {
-                    no_console_log: false,
+                    quiet_level: RNP_QUIET_LEVEL_NONE,
                     csv_log_path: None,
                     json_log_path: None,
                     text_log_path: None,
@@ -353,13 +310,10 @@ mod tests {
     fn parsing_short_options_should_work() {
         assert_eq!(
             RnpCliOptions {
-                common_options: RnpCliCommonOptions {
-                    target: "10.0.0.1:443".parse().unwrap(),
-                    protocol: RnpSupportedProtocol::TCP,
+                common_options: RnpCliCommonOptions { target: "10.0.0.1:443".parse().unwrap(), protocol: RnpSupportedProtocol::TCP },
+                ping_common_options: RnpCliPingCommonOptions {
                     source_ip: "10.0.0.2".parse().unwrap(),
-                    source_ports: Some(PortRangeList {
-                        ranges: vec![(1024..=2048), (3096..=3096), (3097..=3097)]
-                    }),
+                    source_ports: Some(PortRangeList { ranges: vec![(1024..=2048), (3096..=3096), (3097..=3097)] }),
                     ping_count: 10,
                     ping_until_stopped: true,
                     warmup_count: 0,
@@ -367,7 +321,10 @@ mod tests {
                     ping_interval_in_ms: 1500,
                     time_to_live: None,
                     check_disconnect: true,
+                    wait_before_disconnect_in_ms: 0,
+                    disconnect_timeout_in_ms: 1000,
                     parallel_ping_count: 10,
+                    exit_on_fail: false,
                 },
                 quic_options: RnpCliQuicPingOptions {
                     server_name: None,
@@ -376,7 +333,7 @@ mod tests {
                     use_timer_rtt: false,
                 },
                 output_options: RnpCliOutputOptions {
-                    no_console_log: true,
+                    quiet_level: RNP_QUIET_LEVEL_NO_PING_RESULT,
                     csv_log_path: Some(PathBuf::from("log.csv")),
                     json_log_path: Some(PathBuf::from("log.json")),
                     text_log_path: Some(PathBuf::from("log.txt")),
@@ -402,6 +359,8 @@ mod tests {
                 "-i",
                 "1500",
                 "-d",
+                "--dt",
+                "1000",
                 "-p",
                 "10",
                 "-q",
@@ -423,13 +382,10 @@ mod tests {
     fn parsing_long_options_should_work() {
         assert_eq!(
             RnpCliOptions {
-                common_options: RnpCliCommonOptions {
-                    target: "10.0.0.1:443".parse().unwrap(),
-                    protocol: RnpSupportedProtocol::QUIC,
+                common_options: RnpCliCommonOptions { target: "10.0.0.1:443".parse().unwrap(), protocol: RnpSupportedProtocol::QUIC },
+                ping_common_options: RnpCliPingCommonOptions {
                     source_ip: "10.0.0.2".parse().unwrap(),
-                    source_ports: Some(PortRangeList {
-                        ranges: vec![(1024..=2048), (3096..=3096), (3097..=3097)]
-                    }),
+                    source_ports: Some(PortRangeList { ranges: vec![(1024..=2048), (3096..=3096), (3097..=3097)] }),
                     ping_count: 10,
                     ping_until_stopped: false,
                     warmup_count: 3,
@@ -437,7 +393,10 @@ mod tests {
                     ping_interval_in_ms: 1500,
                     time_to_live: Some(128),
                     check_disconnect: true,
+                    wait_before_disconnect_in_ms: 3000,
+                    disconnect_timeout_in_ms: 4000,
                     parallel_ping_count: 10,
+                    exit_on_fail: true,
                 },
                 quic_options: RnpCliQuicPingOptions {
                     server_name: Some(String::from("localhost")),
@@ -446,7 +405,7 @@ mod tests {
                     use_timer_rtt: true,
                 },
                 output_options: RnpCliOutputOptions {
-                    no_console_log: true,
+                    quiet_level: RNP_QUIET_LEVEL_NO_OUTPUT,
                     csv_log_path: Some(PathBuf::from("log.csv")),
                     json_log_path: Some(PathBuf::from("log.json")),
                     text_log_path: Some(PathBuf::from("log.txt")),
@@ -475,15 +434,20 @@ mod tests {
                 "--ttl",
                 "128",
                 "--check-disconnect",
+                "--wait-before-disconnect",
+                "3000",
+                "--disconnect-timeout",
+                "4000",
                 "--parallel",
                 "10",
+                "--exit-on-fail",
                 "--server-name",
                 "localhost",
                 "--log-tls-key",
                 "--alpn",
                 "hq-29",
                 "--use-timer-rtt",
-                "--no-console-log",
+                "-qqq",
                 "--log-csv",
                 "log.csv",
                 "--log-json",
@@ -501,7 +465,7 @@ mod tests {
     #[test]
     fn new_from_cli_options_should_work() {
         assert_eq!(
-            RnpCoreConfig {
+            RnpPingRunnerConfig {
                 worker_config: PingWorkerConfig {
                     protocol: RnpSupportedProtocol::TCP,
                     target: "10.0.0.1:443".parse().unwrap(),
@@ -511,6 +475,8 @@ mod tests {
                         wait_timeout: Duration::from_millis(1000),
                         time_to_live: Some(128),
                         check_disconnect: false,
+                        wait_before_disconnect: Duration::from_millis(2000),
+                        disconnect_timeout: Duration::from_millis(3000),
                         server_name: None,
                         log_tls_key: false,
                         alpn_protocol: None,
@@ -518,15 +484,15 @@ mod tests {
                     },
                 },
                 worker_scheduler_config: PingWorkerSchedulerConfig {
-                    source_ports: PortRangeList {
-                        ranges: vec![(1024..=2048), (3096..=3096), (3097..=3097)]
-                    },
+                    source_ports: PortRangeList { ranges: vec![(1024..=2048), (3096..=3096), (3097..=3097)] },
                     ping_count: Some(4),
                     warmup_count: 1,
                     parallel_ping_count: 1,
                 },
                 result_processor_config: PingResultProcessorConfig {
-                    no_console_log: false,
+                    common_config: PingResultProcessorCommonConfig { quiet_level: RNP_QUIET_LEVEL_NONE },
+                    exit_on_fail: false,
+                    exit_failure_reason: None,
                     csv_log_path: None,
                     json_log_path: None,
                     text_log_path: None,
@@ -538,21 +504,21 @@ mod tests {
                 extra_ping_result_processors: vec![],
             },
             RnpCliOptions {
-                common_options: RnpCliCommonOptions {
-                    target: "10.0.0.1:443".parse().unwrap(),
-                    protocol: RnpSupportedProtocol::TCP,
+                common_options: RnpCliCommonOptions { target: "10.0.0.1:443".parse().unwrap(), protocol: RnpSupportedProtocol::TCP },
+                ping_common_options: RnpCliPingCommonOptions {
                     ping_count: 4,
                     ping_until_stopped: false,
                     warmup_count: 1,
                     source_ip: "10.0.0.2".parse().unwrap(),
-                    source_ports: Some(PortRangeList {
-                        ranges: vec![(1024..=2048), (3096..=3096), (3097..=3097)]
-                    }),
+                    source_ports: Some(PortRangeList { ranges: vec![(1024..=2048), (3096..=3096), (3097..=3097)] }),
                     wait_timeout_in_ms: 1000,
                     ping_interval_in_ms: 1500,
                     time_to_live: Some(128),
                     check_disconnect: false,
+                    wait_before_disconnect_in_ms: 2000,
+                    disconnect_timeout_in_ms: 3000,
                     parallel_ping_count: 1,
+                    exit_on_fail: false,
                 },
                 quic_options: RnpCliQuicPingOptions {
                     server_name: None,
@@ -561,7 +527,7 @@ mod tests {
                     use_timer_rtt: false,
                 },
                 output_options: RnpCliOutputOptions {
-                    no_console_log: false,
+                    quiet_level: RNP_QUIET_LEVEL_NONE,
                     csv_log_path: None,
                     json_log_path: None,
                     text_log_path: None,
@@ -570,11 +536,11 @@ mod tests {
                     latency_buckets: None,
                 },
             }
-            .to_rnp_core_config()
+            .to_ping_runner_config()
         );
 
         assert_eq!(
-            RnpCoreConfig {
+            RnpPingRunnerConfig {
                 worker_config: PingWorkerConfig {
                     protocol: RnpSupportedProtocol::QUIC,
                     target: "10.0.0.1:443".parse().unwrap(),
@@ -584,6 +550,8 @@ mod tests {
                         wait_timeout: Duration::from_millis(2000),
                         time_to_live: Some(128),
                         check_disconnect: true,
+                        wait_before_disconnect: Duration::from_millis(3000),
+                        disconnect_timeout: Duration::from_millis(4000),
                         server_name: Some(String::from("localhost")),
                         log_tls_key: true,
                         alpn_protocol: Some(String::from("h3")),
@@ -591,15 +559,15 @@ mod tests {
                     },
                 },
                 worker_scheduler_config: PingWorkerSchedulerConfig {
-                    source_ports: PortRangeList {
-                        ranges: vec![(1024..=2048), (3096..=3096), (3097..=3097)]
-                    },
+                    source_ports: PortRangeList { ranges: vec![(1024..=2048), (3096..=3096), (3097..=3097)] },
                     ping_count: None,
                     warmup_count: 3,
                     parallel_ping_count: 1,
                 },
                 result_processor_config: PingResultProcessorConfig {
-                    no_console_log: true,
+                    common_config: PingResultProcessorCommonConfig { quiet_level: RNP_QUIET_LEVEL_NO_PING_RESULT },
+                    exit_on_fail: true,
+                    exit_failure_reason: Some(Arc::new(Mutex::new(None))),
                     csv_log_path: Some(PathBuf::from("log.csv")),
                     json_log_path: Some(PathBuf::from("log.json")),
                     text_log_path: Some(PathBuf::from("log.txt")),
@@ -611,21 +579,21 @@ mod tests {
                 extra_ping_result_processors: vec![],
             },
             RnpCliOptions {
-                common_options: RnpCliCommonOptions {
-                    target: "10.0.0.1:443".parse().unwrap(),
-                    protocol: RnpSupportedProtocol::QUIC,
+                common_options: RnpCliCommonOptions { target: "10.0.0.1:443".parse().unwrap(), protocol: RnpSupportedProtocol::QUIC },
+                ping_common_options: RnpCliPingCommonOptions {
                     ping_count: 4,
                     ping_until_stopped: true,
                     warmup_count: 3,
                     source_ip: "10.0.0.2".parse().unwrap(),
-                    source_ports: Some(PortRangeList {
-                        ranges: vec![(1024..=2048), (3096..=3096), (3097..=3097)]
-                    }),
+                    source_ports: Some(PortRangeList { ranges: vec![(1024..=2048), (3096..=3096), (3097..=3097)] }),
                     wait_timeout_in_ms: 2000,
                     ping_interval_in_ms: 1500,
                     time_to_live: Some(128),
                     check_disconnect: true,
+                    wait_before_disconnect_in_ms: 3000,
+                    disconnect_timeout_in_ms: 4000,
                     parallel_ping_count: 1,
+                    exit_on_fail: true,
                 },
                 quic_options: RnpCliQuicPingOptions {
                     server_name: Some(String::from("localhost")),
@@ -634,7 +602,7 @@ mod tests {
                     use_timer_rtt: true,
                 },
                 output_options: RnpCliOutputOptions {
-                    no_console_log: true,
+                    quiet_level: RNP_QUIET_LEVEL_NO_PING_RESULT,
                     csv_log_path: Some(PathBuf::from("log.csv")),
                     json_log_path: Some(PathBuf::from("log.json")),
                     text_log_path: Some(PathBuf::from("log.txt")),
@@ -643,7 +611,7 @@ mod tests {
                     latency_buckets: Some(vec![0.1, 0.5, 1.0, 10.0]),
                 },
             }
-            .to_rnp_core_config()
+            .to_ping_runner_config()
         );
     }
 
@@ -652,54 +620,37 @@ mod tests {
         let mut opts = RnpCliOptions::from_iter(&["rnp.exe", "10.0.0.1:443"]);
         opts.prepare_to_use();
 
-        assert!(opts.common_options.source_ports.is_some());
-        assert_eq!(
-            1,
-            opts.common_options
-                .source_ports
-                .as_ref()
-                .unwrap()
-                .ranges
-                .len()
-        );
+        assert!(opts.ping_common_options.source_ports.is_some());
+        assert_eq!(1, opts.ping_common_options.source_ports.as_ref().unwrap().ranges.len());
         assert_eq!(
             2000,
-            opts.common_options.source_ports.as_ref().unwrap().ranges[0].end()
-                - opts.common_options.source_ports.as_ref().unwrap().ranges[0].start()
+            opts.ping_common_options.source_ports.as_ref().unwrap().ranges[0].end()
+                - opts.ping_common_options.source_ports.as_ref().unwrap().ranges[0].start()
         );
     }
 
     #[test]
     fn invalid_options_for_ipv4_should_be_fixed() {
         let mut opts = RnpCliOptions::from_iter(&["rnp.exe", "10.0.0.1:443"]);
-        opts.common_options.source_ports = None;
-        opts.common_options.ping_count = 0;
-        opts.common_options.parallel_ping_count = 0;
+        opts.ping_common_options.source_ports = None;
+        opts.ping_common_options.ping_count = 0;
+        opts.ping_common_options.parallel_ping_count = 0;
         opts.output_options.latency_buckets = Some(vec![0.0]);
         opts.prepare_to_use();
 
-        assert_eq!(1, opts.common_options.ping_count);
-        assert_eq!(1, opts.common_options.parallel_ping_count);
-        assert_eq!(
-            Some(vec![
-                0.1, 0.5, 1.0, 5.0, 10.0, 30.0, 50.0, 100.0, 300.0, 500.0
-            ]),
-            opts.output_options.latency_buckets
-        );
+        assert_eq!(1, opts.ping_common_options.ping_count);
+        assert_eq!(1, opts.ping_common_options.parallel_ping_count);
+        assert_eq!(Some(vec![0.1, 0.5, 1.0, 5.0, 10.0, 30.0, 50.0, 100.0, 300.0, 500.0]), opts.output_options.latency_buckets);
 
-        opts.common_options.source_ports = Some(PortRangeList {
-            ranges: vec![(1024..=1047)],
-        });
-        opts.common_options.parallel_ping_count = 100;
+        opts.ping_common_options.source_ports = Some(PortRangeList { ranges: vec![(1024..=1047)] });
+        opts.ping_common_options.parallel_ping_count = 100;
         opts.prepare_to_use();
-        assert_eq!(24, opts.common_options.parallel_ping_count);
+        assert_eq!(24, opts.ping_common_options.parallel_ping_count);
 
-        opts.common_options.source_ports = Some(PortRangeList {
-            ranges: vec![(1024..=1024), (1025..=1025), (1026..=1026)],
-        });
-        opts.common_options.parallel_ping_count = 100;
+        opts.ping_common_options.source_ports = Some(PortRangeList { ranges: vec![(1024..=1024), (1025..=1025), (1026..=1026)] });
+        opts.ping_common_options.parallel_ping_count = 100;
         opts.prepare_to_use();
-        assert_eq!(3, opts.common_options.parallel_ping_count);
+        assert_eq!(3, opts.ping_common_options.parallel_ping_count);
     }
 
     #[test]
@@ -708,7 +659,7 @@ mod tests {
         opts.prepare_to_use();
 
         // If source ip is not set (unspecified/any), we update the IP accordingly to match our target.
-        assert!(opts.common_options.source_ip.is_ipv6());
-        assert_eq!(Ipv6Addr::UNSPECIFIED, opts.common_options.source_ip);
+        assert!(opts.ping_common_options.source_ip.is_ipv6());
+        assert_eq!(Ipv6Addr::UNSPECIFIED, opts.ping_common_options.source_ip);
     }
 }
